//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, C++ rewrite 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  commands.cpp
//! \brief Function fridaCommand: executes one main-level command.

#include "defs.hpp"
#include "ptr.hpp"

#include <cstring>
#include <boost/filesystem.hpp>

#include "../plot/dualplot.hpp"
#include "../plot/plowin.hpp"
#include "../readplus/ask.hpp"
#include "../trivia/file_ops.hpp"

#include "config.hpp"
#include "curve.hpp"
#include "edif.hpp"
#include "expr.hpp"
#include "file_in.hpp"
#include "file_out.hpp"
#include "fit.hpp"
#include "fregistry.hpp"
#include "fsel.hpp"
#include "func.hpp"
#include "import.hpp"
#include "manip.hpp"
#include "mem.hpp"
#include "obj.hpp"
#include "olf.hpp"
#include "opr.hpp"
#include "plot.hpp"
#include "reduce_curv.hpp"
#include "reduce_spec.hpp"
#include "rssm.hpp"
#include "special.hpp"
#include "variables.hpp"

#include "commands.hpp"

namespace fs = boost::filesystem;


//! Executes cmd if is a Frida command, and returns true when successful.

bool frida_command(string cmd)
{
    int i;

    // Extract overwrite marker:
    SMem::instance()->overwrite = 0;
    if (cmd.size() > 1 && cmd[cmd.size() - 1] == '!') {
        SMem::instance()->overwrite = 1;
        cmd.erase(cmd.size() - 1, 1);
    }

    // If there are new files, make them new default
    SFSel::instance()->sel_collect_begin();

    if (cmd == "h" || cmd == "?") {
        cout << "This is Frida, version " VERSION "\n"
                "Plain commands:\n"
                "  a    add to plot\n"
                "  p    plot (g for list of other plot commands)\n"
                "  t    show execution time (for development)\n"
                "  !    execute shell command\n"
                "  quit terminate Frida session\n"
                "Command groups (type initial letter for help):\n"
                "  c    curves and fits\n"
                "  d    directory (inspect internal files)\n"
                "  e    edit (change textual parameters)\n"
                "  g    graphics setup\n"
                "  f    file input, output, creation\n"
                "  m    manipulate data (select, bin, reorganise, ...)\n"
                "  o    operate on variables, compute integrals, ...\n"
                "  r    read raw data (currently only for SPHERES)\n"
                "  _    special / temporary commands\n"
                "Modifier (appended to command):\n"
                "  !    operate in place (overwrite)\n"
                "More help:\n"
                "  hc   constants\n"
                "  hf   functions of scalar arguments\n"
                "  hid  dimension-reducing functionals of data (replaced 'oio')\n"
                "  hic  dimension-reducing functionals of curves\n"
                "  ho   operators by precedence\n"
                "  hr   handling resolution functions (convolution, Dirac delta function)\n"
                "  hp   parameters\n"
                "  \\h   meta commands and macros\n";

    } else if (cmd == "hc") {
        cout << "Constants for use in expressions\n"
                "  e   2.718.. Euler number\n"
                "  pi  3.141.. \n"
                "  inf infinity\n";
    } else if (cmd == "hf") {
        SFuncRegistry::instance()->display_functions();
    } else if (cmd == "hi") {
        cout << "  hid  dimension-reducing functionals of data (replacing 'oio')\n"
                "  hic  dimension-reducing functionals of curves\n";
    } else if (cmd == "hid" || cmd == "oio") {
        SGeniRegistry::instance()->display_entries();
    } else if (cmd == "hic") {
        SCvinRegistry::instance()->display_entries();
    } else if (cmd == "ho") {
        SFuncRegistry::instance()->display_operators();
    } else if (cmd == "hr") {
        cout << "Generic form of fit function that is to be convolved with a resolution spectrum:\n"
                "    p0+p1*resol(p9)+p2*conv(f(t,p3,...,p8),p9)\n"
                "Notes:\n"
                "    - flat background of amplitude p0 does not need convolution\n"
                "    - 'resol' copies the resolution data\n"
                "      (implicit convolution of a Dirac delta distribution with the resolution)\n"
                "    - 'conv' convolutes the given function f(t,...) with the resolution\n"
                "    - p9 is a shift in t; most often it can be set to zero\n"
                "      (a zero shift argument may be omitted from 'conv', but not from 'resol')\n"
                "To activate the convolution:\n"
                "    - use command 'cv' to provide resolution file reference\n"
                "If function f has a sharp peak (strong variation within one resolution bin):\n"
                "    - use 'pconv(F(t,...))', where F is the primitive of f\n";
    } else if (cmd == "hp") {
        cout << "Parameters for use in expressions\n"
                "  t         dummy argument in curve definition\n"
                "  k         index running over files in online memory\n"
                "  j         index running over slices (spectra or curves) in a file\n"
                "  i         index running over points in a spectrum\n"
                "  nk        number of files in online memory\n"
                "  nj[K]     number of slices (spectra or curves) in a file\n"
                "  ni[K,J]   number of points in a spectrum\n"
                "  rL[K]     file-wide parameter number L\n"
                "  zL[K,J]   slice-wide parameter number L\n"
                "  pL[K,J]   curve parameter number L\n"
                "  f[K,J](A) curve's function value for argument A\n"
                "  x[K,J,I]  x-value of a point in a spectrum\n"
                "  y[K,J,I]  y-value of a point in a spectrum\n"
                "  dy[K,J,I] error of y[K,J,I]\n"
                "Note on index placeholders K,J,I:\n"
                "  - each of them can be omitted\n"
                "  - when omitted, the respective default value k, j, or i is used\n"
                "  - no comma needed when only omitted indices follow\n"
                "  - the bracket may also be omitted altogether\n"
                "  - hence are equivalent: x = x[k,j,i] = x[k,j,] = x[k,j] = x[k,] = x[k]\n";

    } else if (cmd == "c") {
        cout << "Curve and fit commands:\n"
                "  cc   create\n"
                "  cca  create artificial curve (no data reference)\n"
                "Basic operations:\n"
                "  cm   modify function\n"
                "  cd   set data file to fit\n"
                "  cv   set file to convolute with (cv-: no convolution)\n"
                "  cf   fit\n"
                "  cfs  fit, allow slow convolution\n"
                "  cp   print parameters\n"
                "Parameter attributes:\n"
                "  cx   fix parameters\n"
                "  cg   make parameters global (same for all curves in the file)\n"
                "  cu   unfix parameters (remove fixed and global attributes)\n"
                "File attributes:\n"
                "  cwc  weighing: constant\n"
                "  cwl  weighing: logarithmic\n"
                "  cwv  weighing: reciprocal variance (default)\n"
                "  cwvd weighing: reciprocal variance, data only\n"
                "  cwvc weighing: reciprocal variance, curve only\n"
                "  cgd+ graphics: take x points from data reference file\n"
                "  cgd- graphics: create x points by dividing the plot window\n"
                "  cr   set restrictions for plot range\n"
                "Frida-wide settings:\n"
                "  cs?  show fit parameters, and indicate commands to modify them\n"
                "  cn?  show parameters for numeric integration, and indicate commands to modify "
                "them\n";

    } else if (cmd == "ci") {
        cout << "obsolete command: 'ci 1' has become 'oi p1'\n";

    } else if (cmd == "cc") {
        NCurveFile::create_fitcurve();
    } else if (cmd == "cca") {
        NCurveFile::create_freecurve();

    } else if (cmd == "cm") {
        NCurveFile::change_expr();

    } else if (cmd == "cd" || cmd == "cv" || cmd == "cv-") {
        NCurveFile::set_file_reference(cmd.substr(1));

    } else if (cmd == "cr") {
        NCurveFile::change_prange();

    } else if (cmd == "cp") {
        NCurveFile::show_pars();

    } else if (cmd == "cf") {
        NCurveFit::fit(false);
    } else if (cmd == "cfs") {
        NCurveFit::fit(true);
    } else if (cmd == "cx" || cmd == "cg" || cmd == "cu") {
        NCurveFile::set_par_attr(cmd[1]);
    } else if (cmd.substr(0, 2) == "cw" || cmd.substr(0, 2) == "cg") {
        NCurveFile::set_properties(cmd.substr(1));

    } else if (cmd.substr(0, 2) == "cs") {
        NCurveFit::set_fit_tuning_pars(cmd.substr(2));

    } else if (cmd.substr(0, 2) == "cn") {
        NCvin::set_integration_tuning_pars(cmd.substr(2));


    } else if (cmd == "d") {
        cout << "Directory commands: inspect internal files\n"
            "  dc   coordinate names and units\n"
            "  dd   doc lines\n"
            "  df   list of files\n"
            "  dp   x|y-values\n"
            "  dr   real parameters\n"
            "  dv   variables\n"
            "  dy   y-values\n"
            "  dz   z-values and spectral ranges\n";

    } else if (cmd == "dc") {
        NEdif::show_coord();
    } else if (cmd == "dd") {
        NEdif::show_doc();
    } else if (cmd == "df") {
        NEdif::show_files();
    } else if (cmd == "dp") {
        NOperate::show("xyd");
    } else if (cmd == "dr") {
        NEdif::show_numpars();
    } else if (cmd == "dv") {
        SVariRegistry::instance()->display_all();
    } else if (cmd == "dy") {
        NOperate::show("y");
    } else if (cmd == "dz") {
        NEdif::show_spectra();

    } else if (cmd == "e") {
        cout << "Edit commands:\n"
                "  ef         file name\n"
                "  ec<coord>  name and unit of <coord>=x,y,z0,..\n"
                "  ed         documentation lines\n"
                "  ev         edit variable\n";

    } else if (cmd.substr(0, 2) == "ec") {
        NEdif::edit_coord(cmd.substr(2));
    } else if (cmd == "ed") {
        NEdif::edit_doc();
    } else if (cmd == "ef") {
        NEdif::edit_fnam();
    } else if (cmd == "ev") {
        NEdif::edit_var();


    } else if (cmd == "f") {
        cout << "File commands:\n"
                "  fl   load from disk\n"
                "  fs   save to disk (use modifier ! to overwrite)\n"
                "  fe1  export (only z and y, one line per spectrum)\n"
                "  fe2  export (one block per spectrum: z header and x-y lines)\n"
                "  fc   copy\n"
                "  fx   exchange (copy, then delete source)\n"
                "  fdel delete\n"
                "  fm   make\n"
                "  ft(h|v|c)[m|b][s][d]  read tab\n"
                "       h horizontal (only y values, one line per spectrum)\n"
                "       v vertical (one x-y pair per line)\n"
                "       c choose vertical columns\n"
                "       m multi-block (blocks begin with z line)\n"
                "       b multi-block (without z lines)\n"
                "       s from script\n"
                "       d input includes dy\n";

    } else if (cmd == "fl") {
        NFileIn::load();
    } else if (cmd == "fs") {
        NFileOut::save("yda", SMem::instance()->overwrite);
    } else if (cmd == "fe1") {
        NFileOut::save("csv", SMem::instance()->overwrite);
    } else if (cmd == "fe2") {
        NFileOut::save("tab", SMem::instance()->overwrite);
    } else if (cmd == "fc") {
        SMem::instance()->mem_copy();
    } else if (cmd == "fx") {
        SMem::instance()->mem_xchange();
    } else if (cmd == "fdel") {
        SFSel::instance()->sel_delete();
    } else if (cmd == "fm") {
        NImport::make_grid();
    } else if (cmd.substr(0, 2) == "ft") {
        NImport::read_tab(cmd.substr(2));


    } else if (cmd == "g") {
        cout << "Graphics commands:\n"
                "  p    plot\n"
                "  pr   plot with new autoranges\n"
                "  pn   plot next spectrum\n"
                "  pv   plot previous spectrum\n"
                "  a    add to plot\n"
                "  an   add next spectrum\n"
                "  av   add previous spectrum\n"
                "  gd   graphic dialog (feed commands to gnuplot)\n"
                "Save for print:\n"
                "  gp   plot to complete .ps file\n"
                "  gf   plot to short .psa file\n"
                "  gfa  append plot to existing .ps or .psa file\n"
                "Change window:\n"
                "  gw   list of plot windows\n"
                "  g<n> switch to plot window <n>\n"
                "Set axes:\n"
                "  gt   2D plot (toggle; or use gt+ gt-)\n"
                "  gx   set x range\n"
                "  gy   set y range\n"
                "  gz   set z range (for 2D plots)\n"
                "  ga   auto-ranges\n"
                "  gxa  x auto-range\n"
                "  gya  y auto-range\n"
                "  gza  z auto-range (for 2D plots)\n"
                "  gxl  log x (toggle; or use gxl+ gxl-)\n"
                "  gyl  log y\n"
                "  gzl  log z\n"
                "  gxf  force x into frame (toggle; or use gxf+ gxf-)\n"
                "  gyf  force y into frame (toggle; or use gyf+ gyf-)\n"
                "Other setup:\n"
                "  ge   plot error bars (toggle, or ge+ ge-)\n"
                "  gnd  set maximum number of data points\n"
                "  gnc  number of equidistant base points for curve plotting\n"
                "  gcr  allow iterative refinement for curve plotting?\n"
                "Help:\n"
                "  gsym explain plot symbols\n";

    } else if (cmd == "gx") {
        SPloWin::instance()->current()->X.ask_and_set("Plot x range");
    } else if (cmd == "gy") {
        SPloWin::instance()->current()->Y.ask_and_set("Plot y range");
    } else if (cmd == "gz") {
        SPloWin::instance()->current()->Z.ask_and_set("Plot z range");
    } else if (cmd == "ga") {
        SPloWin::instance()->current()->X.set_auto();
        SPloWin::instance()->current()->Y.set_auto();
        SPloWin::instance()->current()->Z.set_auto();
    } else if (cmd == "gxa") {
        SPloWin::instance()->current()->X.set_auto();
    } else if (cmd == "gya") {
        SPloWin::instance()->current()->Y.set_auto();
    } else if (cmd == "gza") {
        SPloWin::instance()->current()->Z.set_auto();
    } else if (cmd.substr(0, 2) == "gx" || cmd.substr(0, 2) == "gy" || cmd.substr(0, 2) == "gz"
               || cmd.substr(0, 2) == "ge" || cmd.substr(0, 2) == "gt") {
        SPloWin::instance()->current()->set_aux(cmd);
    } else if (cmd == "gnd") {
        SPloWin::instance()->current()->maxpoints = iask(
            "Maximum number of data points in plot", SPloWin::instance()->current()->maxpoints);
    } else if (cmd == "gnc") {
        SPloWin::instance()->current()->equipoints = iask(
            "Number of equidistant base points for curve plotting",
            SPloWin::instance()->current()->equipoints);
    } else if (cmd == "gcr") {
        SPloWin::instance()->current()->refine = bask(
            "Allow iterative refinement for curve plotting",
            SPloWin::instance()->current()->refine);
    } else if (cmd == "gp" || cmd == "gf" || cmd == "gfa") {
        bool withDefs = cmd.substr(0,2)=="gp";
        const string ext = string(".") + (withDefs ? "ps" : "psa");
        static string fname = "";
        string mode;
        if (cmd != "gfa") {
            fs::path defaultname = triv::next_tmp_file(
                CNode::eval("ps_outdir")->to_s() + (withDefs ? "L" : "l") + "%i" + ext);
            defaultname.replace_extension("");
            fname = wask("Graphic file name (without extension "+ext+")", defaultname.string());
            if (fname=="")
                return false;
            if (fs::path(fname).extension().string()=="")
                fname += ext;
            mode = SMem::instance()->overwrite ? "overwrite" : "create";
        } else { // gfa
            fname = wask("Graphic file name", fname);
            mode = "append";
        }
        if (fname != "")
            SPloWin::instance()->current()->write_postscript(
                triv::wordexp_unique(fname), mode, withDefs,
                SVariRegistry::instance()->to_sMap());
    } else if (cmd == "gw") {
        SPloWin::instance()->display_list();
    } else if (sscanf(cmd.c_str(), "g%i", &i) == 1) {
        SPloWin::instance()->select(i);
    } else if (cmd == "gd") { // graph dialog
        string cmd;
        cout << "entering mygnuplot - to leave, type q\n";
        while (1) {
            cmd = sask("mygnuplot> ");
            if (cmd.substr(0, 1) == "q")
                break;
            SPloWin::instance()->current()->gp_write(cmd);
        }
    } else if (cmd == "gsym") {
        string viewer = CNode::eval("ps_viewer")->to_s();
        string fname = CNode::eval("plot_symbol_legend")->to_s();
        if (!triv::file_exists(fname))
            throw "File "+fname+" not found";
        system((viewer + " " + fname + " >nul 2>nul &").c_str());
    } else if (cmd == "p") {
        NPlot::plot(SPloWin::instance()->current(), false);
    } else if (cmd == "pr") {
        SPloWin::instance()->current()->X.set_auto();
        SPloWin::instance()->current()->Y.set_auto();
        NPlot::plot(SPloWin::instance()->current(), false);
    } else if (cmd == "pn") {
        NPlot::plot(SPloWin::instance()->current(), false, "next");
    } else if (cmd == "pv") {
        NPlot::plot(SPloWin::instance()->current(), false, "prev");
    } else if (cmd == "a") {
        NPlot::plot(SPloWin::instance()->current(), true);
    } else if (cmd == "an") {
        NPlot::plot(SPloWin::instance()->current(), true, "next");
    } else if (cmd == "av") {
        NPlot::plot(SPloWin::instance()->current(), true, "prev");


    } else if (cmd == "k") {
        cout << "Commands to modify the internal file selection:\n"
                "  kn    next\n"
                "  kv    previous\n";
    } else if (cmd == "kn") {
        SFSel::instance()->sel_increment(+1);
    } else if (cmd == "kv") {
        SFSel::instance()->sel_increment(-1);

    } else if (cmd == "m") {
        cout << "Commands to manipulate data:\n"
                "  m/    protect spectra from forthcoming operations\n"
                "  m/-   lift protections\n"
                "Generic manipulations (expression driven):\n"
                "  md    generic delete\n"
                "  mr    generic retain\n"
                "Per point:\n"
                "  mp?   command list\n"
                "Per slice (spectrum/curve):\n"
                "  ms?   command list\n"
                "Per file:\n"
                "  mfj   merge (\"join\")\n"
                "  mfj+  merge, adding z=number-of-input-file\n"
                "  mfjp  merge pointwise\n"
                "  mr-   remove real par\n"
                "Histogram operations:\n"
                "  mh?   command list\n";

    } else if (cmd == "mp?") {
        cout << "Pointwise manipulations:\n"
                "  mpd   delete\n"
                "  mpr   retain\n"
                "  mpa?  average (rebin) command list\n"
                "  mpbf  break into spectra by fixed factor\n"
                "  mpo   sort\n"
                "  mpsym symmetrize: map -x -> x, then average\n"
                "  mpe-  remove error bar\n"
                "  mpgi  interpolate to given x grid\n"
                "  mpgr  redistribute histogram intensity to given x grid\n";

    } else if (cmd == "mpa?") {
        cout << "Commands to average (rebin) adjacent channels:\n"
                "  mpa[d|f|e|r][s]\n"
                "        d only bin channels with duplicate x\n"
                "        f bin size given by fixed factor\n"
                "        e bring absolute error below some bound\n"
                "        r bring relative error below some bound\n"
                "        s error estimate includes sampling error\n";

    } else if (cmd == "ms?") {
        cout << "Pointwise manipulations:\n"
                "  msd   delete\n"
                "  msr   retain\n"
                "  msa   average\n"
                "  msas  average, error estimate includes sampling error\n"
                "  msb   break into files\n"
                "  msj   join\n"
                "  ms*   spawn spectra\n"
                "  msx   z_max <-> x\n"
                "  mso   sort by expression\n"
                "  msoz  sort by z-values\n"
                "  mzx   exchange / rotate z coordinates\n"
                "  mz-   delete z coordinate\n";

    } else if (cmd == "mh?") {
        cout << "Histogram manipulations:\n"
                "  mhm   make histogram: to prepare, print data range\n"
                "  mhma  make histogram: adaptive mode, use data range\n"
                "  mhmf  make histogram: fixed range mode\n";

    } else if (cmd == "m/") {
        NManip::freeze_slices();
    } else if (cmd == "m/-") {
        NManip::unfreeze_all_slices();
    } else if (cmd == "md") {
        NOperate::select(1);
    } else if (cmd == "mr") {
        NOperate::select(0);
    } else if (cmd == "mpd") {
        NManip::points_select(true);
    } else if (cmd == "mpr") {
        NManip::points_select(false);
    } else if (cmd == "mpa") {
        NManip::points_rebin(false);
    } else if (cmd == "mpas") {
        NManip::points_rebin(true);
    } else if (cmd == "mpad") {
        NManip::points_rebin_duplicates(false);
    } else if (cmd == "mpads") {
        NManip::points_rebin_duplicates(true);
    } else if (cmd == "mpaf") {
        NManip::points_rebin_by_factor(false);
    } else if (cmd == "mpafs") {
        NManip::points_rebin_by_factor(true);
    } else if (cmd == "mpae") {
        NManip::points_rebin_by_err(true);
    } else if (cmd == "mpar") {
        NManip::points_rebin_by_err(false);
    } else if (cmd == "mpbf") {
        NManip::points_break_by_factor();
    } else if (cmd == "mpo") {
        NManip::points_sort();
    } else if (cmd == "mpsym") {
        NManip::points_symmetrize();
    } else if (cmd == "mpe-") {
        NManip::points_remove_err();
    } else if (cmd == "mpgi") {
        NManip::points_interpolate();
    } else if (cmd == "mpgr") {
        NManip::points_redistribute();
    } else if (cmd == "msd") {
        NManip::slices_select(true);
    } else if (cmd == "msr") {
        NManip::slices_select(false);
    } else if (cmd == "msa") {
        NManip::slices_rebin(false);
    } else if (cmd == "msas") {
        NManip::slices_rebin(true);
    } else if (cmd == "msb") {
        NManip::slices_break();
    } else if (cmd == "msj") {
        NManip::slices_merge();
    } else if (cmd == "ms*") {
        NManip::slices_spawn();
    } else if (cmd == "msx") {
        NManip::exchange_x_z();
    } else if (cmd == "mzx") {
        NManip::zcoords_reorder();
    } else if (cmd == "mz-") {
        NEdif::show_coordZ();
        NManip::zcoord_delete();
    } else if (cmd == "mso") {
        NManip::slices_sort();
    } else if (cmd == "msoz") {
        NManip::slices_sort_by_z();
    } else if (cmd == "mfjp") {
        NManip::files_merge_pointwise();
    } else if (cmd.substr(0, 3) == "mfj") {
        NManip::files_merge(cmd.substr(3));
    } else if (cmd == "mr-") {
        NEdif::del_numpar();
    } else if (cmd.substr(0, 3) == "mhm") {
        NManip::make_histogram(cmd.substr(3));


    } else if (cmd == "o") {
        cout << "Commands to operate on variables or spectra:\n"
                "  ox       on x\n"
                "  oy       on y\n"
                "  oz0,..   on z0,..\n"
                "  oz+      add new z value\n"
                "  or0,..   on r0,..\n"
                "  or+      add new real par\n"
                "  op0,..   on p0,.. (fit parameter)\n"
                "  op*      on all fit parameters\n"
                "  opa      take absolute value\n"
                "  opc      copy from another curve\n"
                "  ofi      functional integrate\n"
                "  ofs      summation up to current index\n"
                "  ofd      functional derivate\n"
                "  ofac     functional autocorrelate\n"
                "  oi       rank-reducing operations ("
                "generalized integrals"
                ")\n"
                "  oixy     choose two columns, save them as x and y\n"
                "  oixyd    choose three columns, save them as x, y, dy\n";

    } else if (cmd == "op") {
        throw "use op<n> to modify one parameter, or op* to modify all";
    } else if (cmd == "op*") {
        NCurveFile::pars_query();
    } else if (cmd == "opa") {
        NCurveFile::pars_abs();
    } else if (cmd == "opc") {
        NCurveFile::pars_copy();
    } else if (cmd[0] == 'o' && strchr("xydzpr", cmd[1])) {
        NOperate::Pointwise(cmd.substr(1));
    } else if (cmd.substr(0, 2) == "of") {
        NOperate::Functional(cmd.substr(2));
    } else if (cmd == "oi") {
        NOperate::Integral();
    } else if (cmd.substr(0, 4) == "oixy") {
        NOperate::IntXY(cmd.substr(2));


    } else if (cmd == "r") {
        cout << "Read raw data from SPHERES:\n"
                "  ry   load energy spectra\n"
                "  ry2  load also spectra for open state\n"
                "  ry4  load also chopper histograms\n"
                "  ry8  load spectra for two directions separately\n"
                "  ryd  load only spectra for two directions\n"
                "  ryo  load only spectra for open state\n"
                "  ryc  raw counts (no normalization)\n"
                "  rs   load series of energy spectra\n"
                "  rs2  load also spectra for open state\n";

    } else if (cmd == "ry") {
        NRSSM::read_spec(0);
    } else if (cmd == "ry2") {
        NRSSM::read_spec(1);
    } else if (cmd == "ry4") {
        NRSSM::read_spec(3);
    } else if (cmd == "ry8") {
        NRSSM::read_spec(7);
    } else if (cmd == "ryd") {
        NRSSM::read_spec(12);
    } else if (cmd == "ryo") {
        NRSSM::read_spec(9);
    } else if (cmd == "ryc") {
        NRSSM::read_spec(16);
    } else if (cmd == "rs") {
        NRSSM::read_series(0);
    } else if (cmd == "rs2") {
        NRSSM::read_series(1);


    } else if (cmd == "!") {
        string sh_cmd;
        sh_cmd = sask("shell command");
        triv::system(sh_cmd);


    } else if (cmd == "_") {
        cout << "Special programs:\n"
                "  _fc      cosine Fourier transform\n"
                "  _t       test (may change at any moment)\n";


    } else if (cmd == "_fc") {
        NSpecial::FourierCosine();

    } else if (cmd == "_t") {
        NSpecial::Test();

    } else if (cmd == "qui" || cmd == "quit") {
        exit(0);

    } else {
        return false; // command not found
    }

    // If there are new files, make them new default
    SFSel::instance()->sel_collect_end();

    return true;
}
