//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  import.cpp
//! \brief NImport: import tables or make data ex nihilo

#include <boost/filesystem.hpp>

#include "defs.hpp"

#include "../readplus/ask.hpp"
#include "../trivia/file_ops.hpp"
#include "../trivia/string_ops.hpp"

#include "import.hpp"
#include "mem.hpp"
#include "obj.hpp"
#include "olf.hpp"
#include "slice.hpp"

namespace fs = boost::filesystem;


//! Create a new online file from table.

void NImport::read_tab(string qualif)
{ // EMBEDDED_DIALOG
    // ** parse qualif **
    if (qualif.find_first_not_of("hvcsbmd") != string::npos)
        throw S("invalid qualifier");
    char dir = qualif[0];
    if (!(dir == 'h' || dir == 'v' || dir == 'c'))
        throw S("missing qualifier h or v");
    bool horizontal = dir == 'h';
    bool choosecol = dir == 'c';
    static int iycol = 0;
    bool fromscript = qualif.find('s') != string::npos;
    bool multiblock = qualif.find('m') != string::npos;
    bool simpleblock = qualif.find('b') != string::npos;
    if (multiblock && simpleblock)
        throw S("incompatible options bm");
    bool with_d = qualif.find('d') != string::npos;
    if (horizontal && with_d)
        throw S("incompatible options hd");

    // ** input from script or from file **
    static string script;
    vector<string> inFiles;
    if (fromscript) {
        inFiles.push_back("/ram/tab");
        script = sask("Script (writing to " + inFiles[0] + ")", script);
        triv::system(script);
    } else {
        string fnames = wask("Read tab from file(s)");
        inFiles = triv::glob_file_list(fnames, "");
    }

    if (choosecol) {
        printf("at present, x:=i\n");
        iycol = iask("Read y from column", iycol);
        if (iycol < 0)
            return;
    }

    FILE* fd;
    for (int iF = 0; iF < inFiles.size(); ++iF) {
        if (!(fd = fopen(inFiles[iF].c_str(), "r")))
            throw "cannot open file " + inFiles[iF];
        cout << ".. reading from " << inFiles[iF] << "\n";
        POld fout(new COld);
        if (fromscript) {
            fout->log_action("ft" + qualif + " " + script);
            fout->name = script;
        } else {
            fout->log_action("ft" + qualif + " " + inFiles[iF]);
            fout->name = fs::path(inFiles[iF]).filename().stem().string();
            if (choosecol) {
                fout->log_action("y from column " + S(iycol));
                fout->name += "_" + S(iycol);
            }
        }

        // ** file-level settings **
        if (choosecol) {
            fout->xco = CCoord("i", "");
            fout->yco = CCoord("y" + S(iycol), "");
        } else {
            fout->xco = CCoord("x", "");
            fout->yco = CCoord("y", "");
        }

        // *** read input ***
        string lin, s1, s2;
        vector<double> dat, zdat;
        int n_in = -1; // line number
        int nblock = 0;
        int nline = -1; // line number within block (starting out of block)
        int nz = 0;
        int nzdat = 0;
        double val;
        PSpec s;
        while (triv::freadln(fd, &lin)) {
            ++n_in;
            if (lin.substr(0, 4) == "#rpa") {
                triv::string_extract_word(lin.substr(4), &s1, &s2);
                if (!triv::any2dbl(s2, &val))
                    printf("invalid parameter line [%s]\n", lin.c_str());
                else
                    fout->RPar.push_back(CParam(s2, "", val));
                continue;
            }
            if (lin[0] == '#')
                continue;

            if (nline == -1) { // start of block
                if (!horizontal && nblock != 0 && s && s->size() > 0)
                    fout->V.push_back(move(s));
                if (multiblock) {
                    if (!triv::str2vec(lin, &zdat, 0, false))
                        throw "invalid header line [" + lin + "] (or legitimate break ?)";
                    if (nblock == 0) { // first header
                        nzdat = zdat.size();
                        for (int iz = 0; iz < nzdat; ++iz)
                            fout->ZCo.push_back(CCoord("z" + S(iz), ""));
                    } else if (zdat.size() != nzdat)
                        throw "line " + S(n_in) + ": header has length " + S(zdat.size())
                            + " instead of " + S(nzdat);
                } else if (simpleblock) {
                    if (nblock == 0) // first header
                        fout->ZCo.push_back(CCoord("#block", ""));
                    nzdat = 1;
                    zdat = { (double)nblock };
                }
                if (nblock == 0 && horizontal) {
                    nz = 1;
                    if (multiblock || simpleblock)
                        nz += nzdat;
                    fout->ZCo.push_back(CCoord("line", ""));
                }
                if (!horizontal) {
                    s = PSpec(new CSpec);
                    s->z.resize(nzdat);
                    for (int iz = 0; iz < nzdat; ++iz)
                        s->z[iz] = RObjDbl(new CObjDbl(zdat[iz]));
                }
                nline = 0;
                ++nblock;
                if (multiblock) // this was a header line
                    continue;
            }

            if (lin == "" && nline != -1) { // end-of-block
                nline = -1;
                continue;
            }

            // regular data line
            if (!triv::str2vec(lin, &dat, 0, false))
                throw "cannot parse line " + S(nline) + " in block " + S(nblock) + " [" + lin + "]";
            if (dat.size() == 0)
                throw "no data in line " + S(nline) + " in block " + S(nblock) + " [" + lin + "]";

            if (horizontal) { // new spectrum for every line
                s = PSpec(new CSpec);
                s->z.resize(nz);
                for (int iz = 0; iz < nzdat; ++iz)
                    s->z[iz] = RObjDbl(new CObjDbl(zdat[iz]));
                s->z[nz - 1] = RObjInt(new CObjInt(nline));
                for (int i = 0; i < dat.size(); ++i)
                    s->push_xy((double)i, dat[i]);
                fout->V.push_back(move(s));
            } else { // vertical
                if (choosecol) {
                    if (iycol >= dat.size())
                        throw "line " + S(n_in) + " (line " + S(nline) + " in block " + S(nblock)
                            + ") [" + lin + "] contains only " + S(dat.size()) + " values; "
                            + "cannot read y from col " + S(iycol);
                    s->push_xy((double)nline, dat[iycol]);
                } else {
                    if (dat.size() != (with_d ? 3 : 2)) {
                        throw "line " + S(n_in) + " (line " + S(nline) + " in block " + S(nblock)
                            + ") [" + lin + "] contains " + S(dat.size()) + " values; "
                            + " exactly " + S(with_d ? 3 : 2) + " are required";
                    }
                    if (with_d)
                        s->push_xyd(dat[0], dat[1], dat[2]);
                    else
                        s->push_xy(dat[0], dat[1]);
                }
            }
            ++nline;
        } // end of file input loop

        if (!horizontal && s && s->size() > 0)
            fout->V.push_back(move(s));

        if (!(fout->nJ()))
            throw "no input lines";
        if (fout->nJ() == 1) {
            fout->ZCo.clear();
            fout->VS(0)->z.clear();
        }
        SMem::instance()->mem_store(move(fout));
    }
}


//! Create a new online file containing a regular x,z grid; let y=0.

void NImport::make_grid()
{
    POld fout(new COld);

    static int ni = 1, nj = 1;
    static string fnam;

    int niIn = iask("Number of points per spectrum", ni);
    if (niIn < 1)
        return;
    ni = niIn;

    int njIn = iask("Number of spectra", nj);
    if (njIn < 1)
        return;
    nj = njIn;

    fnam = "grid" + S(ni);
    if (nj > 1)
        fnam += "x" + S(nj);

    fout->name = wask("Save as", fnam);
    fout->log_action("fm " + S(ni) + " " + S(nj));

    // *** set coordinates ***
    fout->xco = CCoord("x", "");
    fout->yco = CCoord("y", "");
    if (nj > 1)
        fout->ZCo.push_back(CCoord("z", ""));

    // *** loop over spectra ***
    for (int j = 0; j < nj; ++j) {
        PSpec s(new CSpec);

        s->x.resize(ni);
        if (ni==1) {
            s->x[0] = 0;
        } else {
            for (int i = 0; i < ni; ++i)
                s->x[i] = i / (ni-1.);
        }
        s->y.assign(ni, 0.);

        if (nj > 1)
            s->z.push_back(RObjInt(new CObjInt(j)));

        fout->V.push_back(move(s));
    }

    SMem::instance()->mem_store(move(fout));
}
