//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  obj.hpp
//! \brief Data containers: CObj and its children.

#ifndef OBJ_H
#define OBJ_H

#include "defs.hpp"
#include "ptr.hpp"
#include <cmath>

//! Generic data container, pure virtual base class.

class CObj
{
public:
    virtual ~CObj() {}
    virtual bool is_vec() const = 0;
    virtual bool has_err() const = 0;
    virtual char base_type() const = 0;
    virtual string result_info() const = 0;
    virtual int to_i(int i) const { throw string("to_i(i) not available"); }
    virtual double to_r(int i) const { throw string("to_r(i) not available"); }
    virtual double to_dr(int i) const { throw string("to_dr(i) not available"); }
    virtual int to_i() const { throw string("to_i() not available"); }
    virtual double to_r() const { throw string("to_r() not available"); }
    virtual double to_dr() const { throw string("to_dr() not available"); }
    virtual bool to_b() const = 0;
    virtual string to_s(int maxlen = 12, int minlen = 1, int prec = 6) const = 0;
    virtual int size() const { return 1; }

    static PObj objectify(int val);
    static PObj objectify(double val);
    static PObj objectify(double val, double err);
    static PObj objectify(string val);
};

//! Data container holding a single number; virtual base class for CObjInt, CObjDble.

class CObjNum : public CObj
{
public:
    CObjNum() : CObj() {}
    bool is_vec() const { return false; }
    virtual double to_dr() const { return 0.; }
    virtual double to_dr(int i) const { return 0.; }
    virtual RObjVecNum to_vec(int repetitions) const = 0;
};

//! Data container holding a single integer number.

class CObjInt : public CObjNum
{
public:
    int val; //!< Scalar value.
    CObjInt(int _val) : CObjNum(), val(_val) {}
    inline bool has_err() const { return false; }
    inline char base_type() const { return 'i'; }
    string result_info() const { return "CObjInt(" + S(val) + ")"; }
    inline double to_r() const { return val; }
    inline double to_r(int i) const { return val; }
    inline int to_i() const { return val; }
    inline int to_i(int i) const { return val; }
    string to_s(int maxlen = 12, int minlen = 1, int prec = 6) const;
    inline bool to_b() const { return val; }
    RObjVecNum to_vec(int repetitions) const final;
};

//! Data container holding a single floating-point number.

class CObjDbl : public CObjNum
{
public:
    double val; //!< Scalar value.
    CObjDbl(double _val = NAN) : CObjNum(), val(_val) {}
    inline bool has_err() const { return false; }
    inline char base_type() const { return 'd'; }
    string result_info() const;
    inline double to_r() const { return val; }
    inline double to_r(int i) const { return val; }
    string to_s(int maxlen = 12, int minlen = 1, int prec = 6) const;
    bool to_b() const { return val; }
    virtual RObjVecNum to_vec(int repetitions) const;
};

//! Data container holding a single floating-point number with its error bar.

class CObjEnu : public CObjDbl
{
public:
    double err; //!< Error of scalar value.
    CObjEnu(double _val = NAN, double _err = NAN) : CObjDbl(_val), err(_err) {}
    inline bool has_err() const { return true; }
    inline char base_type() const { return 'e'; }
    string result_info() const;
    inline double to_r() const { return val; }
    inline double to_dr() const { return err; }
    inline double to_r(int i) const { return val; }
    inline double to_dr(int i) const { return err; }
    string to_s(int maxlen = 12, int minlen = 1, int prec = 6) const;
    bool to_b() const { return val; }
    RObjVecNum to_vec(int repetitions) const final;
};

//! Data container holding a string.

class CObjStr : public CObj
{
public:
    string val;
    CObjStr(string _val) : CObj(), val(_val) {}
    bool is_vec() const { return false; }
    bool has_err() const { return false; }
    char base_type() const { return 's'; }
    string result_info() const { return "CObjStr(" + val + ")"; }
    string to_s(int maxlen = 12, int minlen = 1, int prec = 6) const { return val; }
    bool to_b() const { return val.size(); }
    int size() const { return val.size(); }
};

//! Data container holding a vector; virtual base class for different vector types.

class CObjVec : public CObj
{
public:
    CObjVec() : CObj() {}
    bool is_vec() const { return true; }
    virtual int size() const = 0;
    inline bool to_b() const { return size(); }
    virtual RObj to_obj(int i) const = 0;
};

//! Data container holding a vector of objects.

class CObjVecObj : public CObjVec
{
public:
    vector<RObj> v; //!< The data.
    CObjVecObj(int n = 0) : CObjVec(), v(n, nullptr) {}
    inline int size() const { return v.size(); }
    bool has_err() const { throw S("query has_err is ill-defined for generic vector"); }
    char base_type() const { throw S("query base_type is ill-defined for generic vector"); }
    inline string result_info() const { return to_s(); }
    inline RObj to_obj(int i) const { return v[i]; }
    string to_s(int maxlen = 12, int minlen = 1, int prec = 6) const;
    vector<double> to_rvec() const;
    RObj to_vecnum() const;
};

//! Data container holding a vector of numbers.

class CObjVecNum : public CObjVec
{
public:
    CObjVecNum() : CObjVec() {}
    double to_dr(int i) const { return 0; }
};


//! Data container holding a vector of integer numbers.

class CObjVecInt : public CObjVecNum
{
public:
    vector<int> v; //!< The data.
    CObjVecInt(int n=0, int val=0) : CObjVecNum(), v(n, val) {}
    CObjVecInt(const vector<int>& _v) : CObjVecNum(), v(_v) {}
    inline int size() const { return v.size(); }
    bool has_err() const { return false; }
    char base_type() const { return 'i'; }
    string result_info() const;
    inline RObj to_obj(int i) const { return RObjInt(new CObjInt(v[i])); }
    inline int to_i(int i) const { return v[i]; }
    inline double to_r(int i) const { return v[i]; }
    string to_s(int maxlen = 12, int minlen = 1, int prec = 6) const;
};

//! Data container holding a vector of floating-point numbers.

class CObjVecDbl : public CObjVecNum
{
public:
    vector<double> v; //!< Resulting vector.
    CObjVecDbl(int n = 0, double val = NAN) : CObjVecNum(), v(n, val) {}
    CObjVecDbl(const vector<double>& _v) : CObjVecNum(), v(_v) {}
    inline int size() const { return v.size(); }
    bool has_err() const { return false; }
    char base_type() const { return 'd'; }
    string result_info() const;
    inline RObj to_obj(int i) const { return RObjDbl(new CObjDbl(v[i])); }
    inline double to_r(int i) const { return v[i]; }
    string to_s(int maxlen = 12, int minlen = 1, int prec = 6) const;
};

//! Data container holding a vector of floating-point numbers.

class CObjVecEnu : public CObjVecDbl
{
public:
    vector<double> dv; //!< Errors.
    CObjVecEnu(int n = 0) : CObjVecDbl(n), dv(n, NAN) {}
    CObjVecEnu(const vector<double>& _v, const vector<double>& _dv) : CObjVecDbl(_v), dv(_dv) {}
    inline int size() const { return v.size(); }
    bool has_err() const { return true; }
    char base_type() const { return 'e'; }
    string result_info() const;
    inline RObj to_obj(int i) const { return RObjEnu(new CObjEnu(v[i], dv[i])); }
    inline double to_dr(int i) const { return dv[i]; }
    string to_s(int maxlen = 12, int minlen = 1, int prec = 6) const;
};

#endif // OBJ_H
