//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  opr.cpp
//! \brief NOperate: functional operations.

#include "defs.hpp"

#include <boost/format.hpp>

#include "../readplus/ask.hpp"

#include "expr.hpp"
#include "fsel.hpp"
#include "genus.hpp"
#include "loop.hpp"
#include "mem.hpp"
#include "obj.hpp"
#include "olf.hpp"
#include "opr.hpp"
#include "slice.hpp"
#include "xax_lex.hpp"

using boost::format;


//**************************************************************************************************
//*  operate via generic expressions
//**************************************************************************************************


//! Display data. Use boolean expression to display only a subset of the data.

void NOperate::show(const string& subcmd)
{
    FileIterator fiter(SFSel::instance()->selD());

    static string expr = "1";
    expr = wask("Show if", expr);
    if (expr == "")
        return;
    RNode T = user_xaxparse(expr.c_str());
    if (T->has_dummy())
        throw S("dummy argument t not allowed when operating on data");

    while (const COld* fin = fiter.nextD()) {
        if (fiter.size() > 1)
            printf(".. file %i:\n", fiter.k());

        string outj, outp;

        // determine length of longest spectrum:
        int ni = 0;
        for (int j = 0; j < fin->nJ(); j++)
            ni = std::max(ni, fin->VS(j)->size());

        // printout by i:
        for (int i = 0; i < ni; i++) {
            outj = "";
            outp = "";

            for (int j = 0; j < fin->nJ(); j++) {
                const CSpec* s = fin->VS(j);
                if (i >= s->size())
                    continue;
                if (!T->tree_point_dbl(fiter.k(), j, i))
                    continue;
                outj += S(j) + " ";
                if (subcmd == "xyd")
                    if (s->has_dy())
                        outp += str(
                            format("%- 12.5g %- 12.5g +- %- 12.5g") % s->x[i] % s->y[i] % s->dy[i]);
                    else
                        outp += str(format("%- 12.5g %- 12.5g") % s->x[i] % s->y[i]);
                else if (subcmd == "y")
                    outp += str(format("%- 12.5g") % s->y[i]);
                else
                    throw "invalid subcmd " + subcmd;
                outp += " ";
            }

            if (outp == "")
                continue;

            printf(
                "%5i [%s] %s\n", i, outj.substr(0, outj.size() - 1).c_str(),
                outp.substr(0, outp.size() - 1).c_str());
        }
    }
}


//! Delete or retain data points according to a boolean expression.

void NOperate::select(bool sel_del)
{
    FileIterator fiter(SFSel::instance()->sel());
    static string expr;
    expr = wask((sel_del ? "Delete if" : "Retain if"), expr);
    if (expr == "")
        return;
    RNode T = user_xaxparse(expr.c_str());
    if (T->has_dummy())
        throw S("dummy argument t not allowed when operating on data");

    while (const COlo* fin = fiter.next()) {
        const COld* fd = dynamic_cast<const COld*>(fin);
        const COlc* fc = dynamic_cast<const COlc*>(fin);

        POlo fout(fin->new_POlo());
        fout->log_action((sel_del ? "md " : "mr ") + expr);

        for (int j = 0; j < fin->nJ(); j++) {
            if (fd) {
                const CSpec* sin = fd->VS(j);
                PSpec sout = PSpec(new CSpec);
                sout->z = sin->z;
                vector<double> lvec(sin->size());
                T->tree_vec_val(&lvec, 0, fiter.k(), j);
                bool with_dy = sin->has_dy();
                for (int i = 0; i < sin->size(); i++) {
                    if (sel_del ^ (bool)lvec[i]) {
                        if (with_dy)
                            sout->push_xyd(sin->x[i], sin->y[i], sin->dy[i]);
                        else
                            sout->push_xy(sin->x[i], sin->y[i]);
                    }
                }
                if (sout->size())
                    fout->V.push_back(move(sout));
            } else if (fc) {
                PCurve sout = PCurve(new CCurve(*(fc->VC(j))));
                if (sel_del ^ T->tree_point_val(fiter.k(), j)->to_b())
                    fout->V.push_back(move(sout));
            }
        }
        if (!fout->nJ())
            throw S("no data point left");
        SMem::instance()->mem_store(move(fout), fiter.k());
    }
}


//! Pointwise operation on x, y, or z<n>, according to arithmetic expression.

void NOperate::Pointwise(string llabel)
{
    FileIterator fiter(SFSel::instance()->sel());

    CGenus lref(llabel);
    static string expr;
    if (lref.num==-1) {
        if      (llabel[0]=='z')
            expr = wask("Set new z coord" + llabel);
        else if (llabel[0]=='r')
            expr = wask("Set new r parameter" + llabel);
        else
            throw "BUG: unexpected llabel "+llabel;
    } else {
        expr = wask("Set " + llabel, expr);
    }
    if (expr == "")
        return;
    RNode T = user_xaxparse(expr.c_str());
    if (T->has_dummy())
        throw S("dummy argument t not allowed when operating on data");

    while (const COlo* fin = fiter.next()) {
        int k = fiter.k();
        const COld* fd = dynamic_cast<const COld*>(fin);
        const COlc* fc = dynamic_cast<const COlc*>(fin);

        if (lref.pointwise() && !fd)
            throw S("no pointwise operation on curve");
        if (lref.categ == VCateg::CP && !fc)
            continue;

        POlo fout(fin->new_POlo());
        if (lref.categ != VCateg::CP)
            fout->log_action("o" + lref.genus_info() + " " + expr);
        fout->as_on_disk = false;

        CCoord co = T->node_coord(k);
        if (lref.categ == VCateg::X) {
            fout->xco = co;
        } else if (lref.categ == VCateg::Y) {
            fout->yco = co;
        } else if (lref.categ == VCateg::Z) {
            if (lref.num == -1)
                fout->ZCo.push_back(co);
            else if (lref.num >= fin->ZCo.size())
                throw S("no such z coordinate to operate on");
            else
                fout->ZCo[lref.num] = co;
        } else if (lref.categ == VCateg::R) {
            if (lref.num == -1)
                fout->RPar.push_back(CParam(co));
            else if (lref.num >= fin->RPar.size())
                throw S("no such real par to operate on");
            else
                fout->RPar[lref.num].Co = co;
        } else if (lref.categ == VCateg::CP) {
            if (!fc)
                continue; // just ignore op operations on data files
            if (lref.num >= fc->nP)
                throw S("no such parameter to operate on");
            SMem::instance()->overwrite = 1;
            dynamic_cast<COlc*>(fout.get())->clear_fits_outcome();
        }

        if (lref.categ == VCateg::R) {
            fout->copy_mainvec(fin);
            double val = T->tree_point_val(k, 0)->to_r();
            if (lref.num == -1)
                fout->RPar.back().val = val;
            else
                fout->RPar[lref.num].val = val;
        } else {
            fout->V.resize(fin->nJ());
            for (int j = 0; j < fin->nJ(); j++) {
                fout->V[j] = fin->copy_slice(j);
                if (fin->V[j]->frozen)
                    continue;
                if (lref.pointwise()) {
                    vector<double> vout(fd->nPts(j));
                    COld* fdout = dynamic_cast<COld*>(fout.get());
                    if (lref.categ == VCateg::X || lref.categ == VCateg::DY) {
                        T->tree_vec_val(&vout, 0, k, j);
                        if (lref.categ == VCateg::X)
                            fdout->VS(j)->x = move(vout);
                        else if (lref.categ == VCateg::DY)
                            fdout->VS(j)->dy = move(vout);
                        else
                            throw S("BUG: unexpected lref");
                    } else if (lref.categ == VCateg::Y) {
                        vector<double> dvout(fd->nPts(j));
                        T->tree_vec_val(&vout, &dvout, k, j);
                        fdout->VS(j)->y = move(vout);
                        if (dvout.size())
                            fdout->VS(j)->dy = move(dvout);
                        else
                            fdout->VS(j)->dy.clear();
                    } else
                        throw S("BUG: unexpected lref");
                } else if (lref.categ == VCateg::Z) {
                    RObj pret = T->tree_point_val(k, j);
                    if (lref.num == -1)
                        fout->V[j]->z.push_back(pret);
                    else
                        fout->V[j]->z[lref.num] = pret;
                } else if (lref.categ == VCateg::CP) {
                    COlc* fcout = dynamic_cast<COlc*>(fout.get());
                    fcout->VC(j)->P[lref.num] = T->tree_point_val(k, j)->to_r();
                } else {
                    throw S("BUG: invalid oxy");
                }
            }
        }
        SMem::instance()->mem_store(move(fout), fiter.k());
    }
}


//! Integral operation, using arithmetic expression.

void NOperate::Integral()
{
    FileIterator fiter(SFSel::instance()->sel());

    static string expr;
    expr = wask("Functional", expr);
    if (expr == "")
        return;
    RNode T = user_xaxparse(expr.c_str());
    if (T->has_dummy())
        throw S("dummy argument t not allowed when operating on data");

    while (const COlo* fin = fiter.next()) {
        int k = fiter.k();
        int nz = fin->ZCo.size();
        bool savable = nz > 0;

        POld fout;
        PSpec sout;
        if (savable) {
            if (const COlc* fc = dynamic_cast<const COlc*>(fin))
                fout = POld(new COld(fc));
            else
                fout = POld((dynamic_cast<const COld*>(fin))->new_POld());
            fout->log_action("oi " + expr);
            fout->xco = fin->ZCo.back();
            fout->yco = T->node_coord(k);
            fout->ZCo.pop_back();
            sout = PSpec(new CSpec);
            sout->z = fin->V[0]->z;
            sout->z.pop_back();
        }

        for (int j = 0; j < fin->nJ(); j++) {
            if (savable && fin->V[j]->frozen)
                continue;
            double xval;
            RObj res = T->tree_point_val(k, j);
            if (!savable) {
                cout << "result: " << res->to_s() << "\n";
                continue;
            }
            xval = fin->V[j]->z[nz-1]->to_r();
            sout->push_xyd(xval, res->to_r(), res->to_dr());
            if (nz > 1) { // new spectrum if jump in other z values
                if (j+1 < fin->nJ()
		   && fin->V[j+1]->z[nz-2]->to_r() != fin->V[j]->z[nz-2]->to_r()) {
                    fout->V.push_back(move(sout));
                    sout = PSpec(new CSpec);
                    sout->z = fin->V[j+1]->z;
                    sout->z.pop_back();
                }
            }
        }
        if (savable) {
            fout->purge_dy();
            fout->V.push_back(move(sout));
            SMem::instance()->mem_store(move(fout), fiter.k());
        }
        //      if ( !savable )
        //          NReg::lastresult = res;
    }
}


//! Integral operations to generate not only new y, but also new x.

void NOperate::IntXY(string mode)
{ // EMBEDDED_DIALOG
    FileIterator fiter(SFSel::instance()->selD());

    bool with_d;
    if (mode == "xy")
        with_d = false;
    else if (mode == "xyd")
        with_d = true;
    else
        throw S("invalid command");

    static int icolx = 0, icoly = 1, icold = 2;
    static CCoord xco, yco;
    icolx = iask("x from column", icolx);
    if (icolx < 0)
        return;
    icoly = iask("y from column", icoly);
    if (icoly < 0)
        return;
    if (with_d) {
        icold = iask("dy from column", icold);
        if (icold < 0)
            return;
    }

    while (const COld* fin = fiter.nextD()) {
        int nz = fin->ZCo.size();

        POld fout(fin->new_POld());
        string lin = S("oi") + mode + " " + S(icolx) + " " + S(icoly);
        if (with_d)
            lin += " " + S(icold);
        fout->log_action(lin);

        fout->xco = CCoord("x", "");
        fout->yco = CCoord("y", "");
        if (nz)
            fout->ZCo.pop_back();

        PSpec sout(new CSpec);
        sout->z = fin->V[0]->z;
        if (nz)
            sout->z.pop_back();

        for (int j = 0; j < fin->nJ(); j++) {
            int n = fin->nPts(j);
            if (icolx >= n || icoly >= n || (with_d && icold >= n))
                throw "not enough columns in spectrum " + S(j);
            if (with_d)
                sout->push_xyd(fin->VS(j)->y[icolx], fin->VS(j)->y[icoly], fin->VS(j)->y[icold]);
            else
                sout->push_xy(fin->VS(j)->y[icolx], fin->VS(j)->y[icoly]);
            sout->z = fin->VS(j)->z;
            if (nz > 1) { // new spectrum if jump in other z value
                if (j+1 < fin->nJ()
                    && fin->V[j+1]->z[nz-2]->to_r() != fin->V[j]->z[nz-2]->to_r()) {
                    fout->V.push_back(move(sout));
                    sout = PSpec(new CSpec);
                    sout->z = fin->V[j+1]->z;
                    if (nz)
                        sout->z.pop_back();
                }
            }
        }
        fout->V.push_back(move(sout));
        SMem::instance()->mem_store(move(fout), fiter.k());
    }
}

//**************************************************************************************************
//*  functional operations
//**************************************************************************************************

//! Functional integration, derivation, ...

void NOperate::Functional(const string& subcmd)
{
    FileIterator fiter(SFSel::instance()->selD());
    while (const COld* fin = fiter.nextD()) {
        POld fout(fin->new_POld());
        fout->log_action("of" + subcmd);

        if (subcmd == "i") { // integrate
            fout->yco.name = "Int d" + fin->xco.name + " " + fin->yco.name;
            fout->yco.unit = fin->xco.unit + " * " + fin->yco.unit;
        } else if (subcmd == "s") { // sum
            fout->yco.name = "Sum_0^i " + fin->yco.name;
            fout->yco.unit = fin->yco.unit;
        } else if (subcmd == "d") { // derive
            fout->yco.name = "d" + fin->yco.name + " / d" + fin->xco.name;
            fout->yco.unit = fin->yco.unit + " / " + fin->xco.unit;
        } else if (subcmd == "ac") { // autocorrelate
            fout->yco.name = "ac{" + fin->yco.name + "}";
            fout->yco.unit = fin->yco.unit + "^2";
        } else {
            throw S("invalid 'of' command");
        }

        for (int j = 0; j < fin->nJ(); j++) {
            const CSpec* sin = fin->VS(j);
            PSpec sout(new CSpec);
            sout->z = sin->z;
            int n = sin->size();

            if (subcmd == "i") { // integrate
                sout->x = sin->x;
                sout->y.resize(n);
                sout->y[0] = 0;
                for (int i = 1; i < n; ++i)
                    sout->y[i] = sout->y[i - 1]
                        + (sin->x[i] - sin->x[i - 1]) * (sin->y[i] + sin->y[i - 1]) / 2;
            } else if (subcmd == "s") { // sum
                sout->x = sin->x;
                sout->y.resize(n);
                sout->y[0] = sin->y[0];
                for (int i = 1; i < n; ++i)
                    sout->y[i] = sout->y[i - 1] + sin->y[i];
            } else if (subcmd == "d") { // derive
                for (int i = 0; i < n - 1; ++i) {
                    double dx = sin->x[i + 1] - sin->x[i];
                    sout->push_xy(
                        (sin->x[i + 1] + sin->x[i]) / 2,
                        dx == 0 ? 0 : (sin->y[i + 1] - sin->y[i]) / dx);
                }
            } else if (subcmd == "ac") { // autocorrelate
                throw "currently disabled, old code in old/src-fragm/opr-func-ac.cpp";
            }
            fout->V.push_back(move(sout));
        }
        SMem::instance()->mem_store(move(fout), fiter.k());
    }
}
