//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  gnuplotter.cpp
//! \brief Implements class CGnuPlotter.

#include <cmath>
#include <cstring>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <boost/format.hpp>

#include "../trivia/file_ops.hpp"
#include "../trivia/string_convs.hpp"

#include "axis.hpp"
#include "gnuplotter.hpp"

#define S(a) triv::strg((a))

using std::string;
using std::vector;
using boost::format;


CGnuPlotter::CGnuPlotter(int _iPlot)
    : iPlot(_iPlot)
{
    // Create a named pipe (FIFO) that will transmit our commands to Gnuplot.
    string fn_fifo = "/tmp/gnuplot-" + S(getpid());
    triv::system("rm -f " + fn_fifo);
    if (mkfifo(fn_fifo.c_str(), 0666))
        throw "SYSTEM ERROR cannot make fifo " + fn_fifo + ": will not be able to print";

    // Check that Gnuplot supports X11.
    string out = triv::system_read("gnuplot -e 'help X11' 2>&1 < /dev/null");
    if (out.substr(0, 5) == "Sorry" || out.length() < 80)
        throw "Gnuplot seems not to support X11";

    // Start a Gnuplot that will read from our pipe.
    triv::system("gnuplot -title " + S(iPlot) + " -noraise < " + fn_fifo + "&");

    // Open our pipe so that we can write commands to it
    //   (we use 'open' instead of 'fopen' or 'ofstream', because we need non-blocking mode).
    if ((gp_fifo = open(fn_fifo.c_str(), O_WRONLY)) == -1)
        throw "SYSTEM ERROR: cannot open fifo " + fn_fifo + " (" + strerror(errno) + ")";
    fcntl(gp_fifo, F_SETFL, O_NONBLOCK);

    // Now the initialization _within_ Gnuplot.
    write("set terminal x11");
}

//! Send one line to gnuplot fifo.

void CGnuPlotter::write(const string& in)
{
    string out = in + "\n";
    if (::write(gp_fifo, out.c_str(), out.size()) <= 0)
        throw S("could not write to gp_fifo");
}

void CGnuPlotter::start_frame(const CAxis& _X, const CAxis& _Y)
{
    X = &_X;
    Y = &_Y;
    gp_fno = 0;
    gp_fnames = "";
    write("set nologscale");
    string whichlog = "";
    if (X->logflag)
        whichlog += "x";
    if (Y->logflag)
        whichlog += "y";
    if (whichlog != "")
        write("set logscale " + whichlog);
}

void CGnuPlotter::add_spec(
    bool as_line, bool new_slice, bool plot_errorbars, int style_no,
    const vector<double>& xp, const vector<double>& yp, const vector<double>& dyp,
    const vector<string>& zentries,
    const string& xco, const string& yco, const string& info)
{
    static vector<int> ccolor = { 0x880000, 0x008800, 0x000088, 0x006666, 0x660066, 0x666600 };
    static vector<int> pcolor = { 0x5050ff, 0xff0000, 0x00a000, 0xff00ff, 0x00a0c0, 0xa0522d,
                                  0xffa500, 0xff7f50, 0x40d020 };
    // Checks:
    int np = xp.size();
    if (!np)
        throw S("invalid call to CPLot::addSpec: no data points");
    if (np != yp.size())
        throw S("invalid call to CPLot::addSpec: x.size<>y.size");

    // Prepare for live display, to be shown by showSpecs():
    string gp_fnam = str(format("/tmp/%s-%i-%03i.gnu") % getenv("LOGNAME") % iPlot % gp_fno++);
    if (gp_fnames != "")
        gp_fnames += ", ";
    gp_fnames += "\"" + gp_fnam + "\" notitle";
    if (as_line)
        gp_fnames += str(format(" with lines lt 1 lc rgb \"#%6x\"")
                         % ccolor[(style_no-1) % ccolor.size()]);
    else {
        gp_fnames += str(format(" with %s pt %i lc rgb \"#%6x\" lw 2")
                         % (plot_errorbars && dyp.size() ? "yerrorbars" : "points")
                         % style_no % pcolor[(style_no-1) %  pcolor.size()]);
    }
    FILE* gp_fd;
    if (!(gp_fd = fopen(gp_fnam.c_str(), "w")))
        throw "cannot save gnuplot data to " + gp_fnam;
    int nout = 0;
    try {
        for (int i = 0; i < np; i++) {
            if (std::isinf(xp[i]) || std::isinf(yp[i]))
                throw "Data point number "+S(i)+" is invalid: x="+S(xp[i])+", y="+S(yp[i]);
            if (xp[i] < X->inf || xp[i] > X->sup)
                throw "CPlot::addSpec: x["+S(i)+"]="+S(xp[i])+" out of range";
            if (yp[i] < Y->inf || yp[i] > Y->sup)
                throw "CPlot::addSpec: y["+S(i)+"]="+S(yp[i])+" out of range";
            if (plot_errorbars && dyp.size())
                fprintf(gp_fd, "%20.13g %20.13g %20.13g\n", xp[i], yp[i], dyp[i]);
            else
                fprintf(gp_fd, "%20.13g %20.13g\n", xp[i], yp[i]);
            nout++;
        }
    } catch (string& s) {
        fclose(gp_fd);
        throw s;
    } catch (...) {
        fclose(gp_fd);
        throw "BUG: unexpected exception type";
    }
    fclose(gp_fd);
    if (!nout)
        throw "no points in frame: " + info;
}

//! Live display as prepared by addSpec.

void CGnuPlotter::show_specs()
{
    if (gp_fnames == "")
        return;
    write(str(format("plot [%24.16g:%24.16g] [%24.16g:%24.16g] ") %
              X->inf % X->sup % Y->inf % Y->sup)
          + gp_fnames);
}
